@@ -4,14 +4,14 @@
  *
  * Heavily modified by Manuel Novoa III       Mar 12, 2001
  *
- * Version:     $Id: inet_common.h,v 1.1.1.2 2006/02/25 01:18:09 dgel Exp $
+ * Version:     $Id: inet_common.h,v 1.1.1.3 2006/08/28 04:07:22 dgel Exp $
  *
  */
 
 #include <features.h>
 #include <sys/types.h>
 #include <sys/socket.h>
-#include <arpa/inet.h>
+#include "platform.h"
 
 
 extern const char bb_INET_default[];    /* = "default" */
