@@ -1,7 +1,8 @@
 /*
  * Copyright (c) 2003-2006 Erez Zadok
  * Copyright (c) 2003-2006 Charles P. Wright
- * Copyright (c) 2005-2006 Josef Sipek
+ * Copyright (c) 2005-2006 Josef 'Jeff' Sipek
+ * Copyright (c) 2005-2006 Junjiro Okajima
  * Copyright (c) 2005      Arun M. Krishnakumar
  * Copyright (c) 2005-2006 David P. Quigley
  * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
@@ -16,7 +17,7 @@
  * This Copyright notice must be kept intact and distributed with all sources.
  */
 /*
- *  $Id: inode.c,v 1.263 2006/02/09 03:37:42 jsipek Exp $
+ *  $Id: inode.c,v 1.273 2006/08/05 01:28:46 jro Exp $
  */
 
 #include "unionfs.h"
@@ -26,12 +27,8 @@ extern struct dentry *unionfs_lookup(struct inode *, struct dentry *,
 				     struct nameidata *);
 extern int unionfs_readlink(struct dentry *dentry, char __user * buf,
 			    int bufsiz);
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
-extern void unionfs_put_link(struct dentry *dentry, struct nameidata *nd);
-#else
 extern void unionfs_put_link(struct dentry *dentry, struct nameidata *nd,
 			     void *cookie);
-#endif
 
 static int unionfs_create(struct inode *parent, struct dentry *dentry,
 			  int mode, struct nameidata *nd)
@@ -46,7 +43,7 @@ static int unionfs_create(struct inode *parent, struct dentry *dentry,
 
 	print_entry_location();
 	lock_dentry(dentry);
-	fist_print_dentry("IN unionfs_create", dentry);
+	print_dentry("IN unionfs_create", dentry);
 
 	/* We start out in the leftmost branch. */
 	bstart = dbstart(dentry);
@@ -81,7 +78,7 @@ static int unionfs_create(struct inode *parent, struct dentry *dentry,
 		struct dentry *hidden_dir_dentry;
 		struct iattr newattrs;
 
-		down(&whiteout_dentry->d_inode->i_sem);
+		mutex_lock(&whiteout_dentry->d_inode->i_mutex);
 		newattrs.ia_valid = ATTR_CTIME | ATTR_MODE | ATTR_ATIME
 		    | ATTR_MTIME | ATTR_UID | ATTR_GID | ATTR_FORCE
 		    | ATTR_KILL_SUID | ATTR_KILL_SGID;
@@ -97,7 +94,7 @@ static int unionfs_create(struct inode *parent, struct dentry *dentry,
 
 		err = notify_change(whiteout_dentry, &newattrs);
 
-		up(&whiteout_dentry->d_inode->i_sem);
+		mutex_unlock(&whiteout_dentry->d_inode->i_mutex);
 
 		if (err)
 			printk(KERN_WARNING
@@ -168,7 +165,7 @@ static int unionfs_create(struct inode *parent, struct dentry *dentry,
 			}
 		}
 
-		fist_checkinode(parent, "unionfs_create");
+		checkinode(parent, "unionfs_create");
 
 		hidden_parent_dentry = lock_parent(hidden_dentry);
 		if (IS_ERR(hidden_parent_dentry)) {
@@ -205,7 +202,7 @@ static int unionfs_create(struct inode *parent, struct dentry *dentry,
 	DPUT(whiteout_dentry);
 	KFREE(name);
 
-	fist_print_dentry("OUT unionfs_create :", dentry);
+	print_dentry("OUT unionfs_create :", dentry);
 	unlock_dentry(dentry);
 	print_exit_status(err);
 	return err;
@@ -357,7 +354,7 @@ static int unionfs_symlink(struct inode *dir, struct dentry *dentry,
 
 	print_entry_location();
 	lock_dentry(dentry);
-	fist_print_dentry("IN unionfs_symlink", dentry);
+	print_dentry("IN unionfs_symlink", dentry);
 
 	/* We start out in the leftmost branch. */
 	bstart = dbstart(dentry);
@@ -386,8 +383,8 @@ static int unionfs_symlink(struct inode *dir, struct dentry *dentry,
 		/* found a .wh.foo entry, unlink it and then call vfs_symlink() */
 		hidden_dir_dentry = lock_parent(whiteout_dentry);
 
-		fist_print_generic_dentry("HDD", hidden_dir_dentry);
-		fist_print_generic_dentry("WD", whiteout_dentry);
+		print_dentry("HDD", hidden_dir_dentry);
+		print_dentry("WD", whiteout_dentry);
 
 		if (!(err = is_robranch_super(dentry->d_sb, bstart))) {
 			err =
@@ -425,7 +422,7 @@ static int unionfs_symlink(struct inode *dir, struct dentry *dentry,
 				if (IS_ERR(hidden_dentry)) {
 					err = PTR_ERR(hidden_dentry);
 				}
-				fist_dprint(8,
+				dprint(PRINT_DEBUG,
 					    "hidden dentry NULL (or error) for bindex = %d\n",
 					    bindex);
 				continue;
@@ -464,7 +461,7 @@ static int unionfs_symlink(struct inode *dir, struct dentry *dentry,
 		d_drop(dentry);
 
 	KFREE(name);
-	fist_print_dentry("OUT unionfs_symlink :", dentry);
+	print_dentry("OUT unionfs_symlink :", dentry);
 	unlock_dentry(dentry);
 	print_exit_status(err);
 	return err;
@@ -483,7 +480,7 @@ static int unionfs_mkdir(struct inode *parent, struct dentry *dentry, int mode)
 
 	print_entry_location();
 	lock_dentry(dentry);
-	fist_print_dentry("IN unionfs_mkdir", dentry);
+	print_dentry("IN unionfs_mkdir", dentry);
 	bstart = dbstart(dentry);
 
 	hidden_dentry = dtohd(dentry);
@@ -541,7 +538,7 @@ static int unionfs_mkdir(struct inode *parent, struct dentry *dentry, int mode)
 		if (!hidden_dentry) {
 			hidden_dentry = create_parents(parent, dentry, bindex);
 			if (!hidden_dentry || IS_ERR(hidden_dentry)) {
-				fist_dprint(8,
+				dprint(PRINT_DEBUG,
 					    "hidden dentry NULL for bindex = %d\n",
 					    bindex);
 				continue;
@@ -568,6 +565,7 @@ static int unionfs_mkdir(struct inode *parent, struct dentry *dentry, int mode)
 		} else {
 			int i;
 			int bend = dbend(dentry);
+
 			for (i = bindex + 1; i < bend; i++) {
 				if (dtohd_index(dentry, i)) {
 					DPUT(dtohd_index(dentry, i));
@@ -585,23 +583,10 @@ static int unionfs_mkdir(struct inode *parent, struct dentry *dentry, int mode)
 				/* update number of links on parent directory */
 				parent->i_nlink = get_nlinks(parent);
 			}
-			whiteout_dentry = LOOKUP_ONE_LEN(UNIONFS_DIR_OPAQUE,
-							 hidden_dentry,
-							 sizeof
-							 (UNIONFS_DIR_OPAQUE) -
-							 1);
-			if (IS_ERR(whiteout_dentry)) {
-				err = PTR_ERR(whiteout_dentry);
-				goto out;
-			}
-			down(&hidden_dentry->d_inode->i_sem);
-			err = vfs_create(hidden_dentry->d_inode,
-					 whiteout_dentry, 0600, NULL);
-			up(&hidden_dentry->d_inode->i_sem);
-			DPUT(whiteout_dentry);
 
+			err = make_dir_opaque(dentry, dbstart(dentry));
 			if (err) {
-				fist_dprint(8,
+				dprint(PRINT_DEBUG,
 					    "mkdir: error creating directory override entry: %d\n",
 					    err);
 				goto out;
@@ -616,7 +601,7 @@ static int unionfs_mkdir(struct inode *parent, struct dentry *dentry, int mode)
 
 	KFREE(name);
 
-	fist_print_dentry("OUT unionfs_mkdir :", dentry);
+	print_dentry("OUT unionfs_mkdir :", dentry);
 	unlock_dentry(dentry);
 	print_exit_status(err);
 	return err;
@@ -634,7 +619,7 @@ static int unionfs_mknod(struct inode *dir, struct dentry *dentry, int mode,
 
 	print_entry_location();
 	lock_dentry(dentry);
-	fist_print_dentry("IN unionfs_mknod", dentry);
+	print_dentry("IN unionfs_mknod", dentry);
 	bstart = dbstart(dentry);
 
 	hidden_dentry = dtohd(dentry);
@@ -684,7 +669,7 @@ static int unionfs_mknod(struct inode *dir, struct dentry *dentry, int mode,
 		if (!hidden_dentry) {
 			hidden_dentry = create_parents(dir, dentry, bindex);
 			if (!hidden_dentry || IS_ERR(hidden_dentry)) {
-				fist_dprint(8,
+				dprint(PRINT_DEBUG,
 					    "hidden dentry NULL for bindex = %d\n",
 					    bindex);
 				continue;
@@ -729,7 +714,7 @@ static int unionfs_mknod(struct inode *dir, struct dentry *dentry, int mode,
 		KFREE(name);
 	}
 
-	fist_print_dentry("OUT unionfs_mknod :", dentry);
+	print_dentry("OUT unionfs_mknod :", dentry);
 	unlock_dentry(dentry);
 	print_exit_status(err);
 	return err;
@@ -743,7 +728,7 @@ int unionfs_readlink(struct dentry *dentry, char __user * buf, int bufsiz)
 	print_entry_location();
 	lock_dentry(dentry);
 	hidden_dentry = dtohd(dentry);
-	fist_print_dentry("unionfs_readlink IN", dentry);
+	print_dentry("unionfs_readlink IN", dentry);
 
 	if (!hidden_dentry->d_inode->i_op ||
 	    !hidden_dentry->d_inode->i_op->readlink) {
@@ -763,11 +748,7 @@ int unionfs_readlink(struct dentry *dentry, char __user * buf, int bufsiz)
 }
 
 /* We don't lock the dentry here, because readlink does the heavy lifting. */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
-static int unionfs_follow_link(struct dentry *dentry, struct nameidata *nd)
-#else
 static void *unionfs_follow_link(struct dentry *dentry, struct nameidata *nd)
-#endif
 {
 	char *buf;
 	int len = PAGE_SIZE, err;
@@ -798,18 +779,10 @@ static void *unionfs_follow_link(struct dentry *dentry, struct nameidata *nd)
 
       out:
 	print_exit_status(err);
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
-	return err;
-#else
 	return ERR_PTR(err);
-#endif
 }
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
-void unionfs_put_link(struct dentry *dentry, struct nameidata *nd)
-#else
 void unionfs_put_link(struct dentry *dentry, struct nameidata *nd, void *cookie)
-#endif
 {
 	char *link;
 	print_entry_location();
@@ -853,23 +826,17 @@ static int inode_permission(struct inode *inode, int mask, struct nameidata *nd,
 		    (!strcmp("nfs", (inode)->i_sb->s_type->name)) &&
 		    (nd) && (nd->mnt) && (nd->mnt->mnt_sb) &&
 		    (branchperms(nd->mnt->mnt_sb, bindex) & MAY_NFSRO)) {
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
-			retval = vfs_permission(inode, submask);
-#else
 			retval = generic_permission(inode, submask, NULL);
-#endif
 		}
 	} else {
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
-		retval = vfs_permission(inode, submask);
-#else
 		retval = generic_permission(inode, submask, NULL);
-#endif
 	}
-	if (retval)
+
+	if (retval && retval != -EROFS) /* ignore EROFS */
 		return retval;
 
-	return security_inode_permission(inode, mask, nd);
+	retval = security_inode_permission(inode, mask, nd);
+	return ((retval == -EROFS) ? 0 : retval); /* ignore EROFS */
 }
 
 static int unionfs_permission(struct inode *inode, int mask,
@@ -878,18 +845,15 @@ static int unionfs_permission(struct inode *inode, int mask,
 	struct inode *hidden_inode = NULL;
 	int err = 0;
 	int bindex, bstart, bend;
-	int is_file = 0;
+	const int is_file = !S_ISDIR(inode->i_mode);
+	const int write_mask = (mask & MAY_WRITE) && !(mask & MAY_READ);
 
 	print_entry_location();
 
 	bstart = ibstart(inode);
 	bend = ibend(inode);
 
-	fist_print_inode("IN unionfs_permission", inode);
-
-	/* set if check is for file */
-	if (!S_ISDIR(inode->i_mode))
-		is_file = 1;
+	print_inode("IN unionfs_permission", inode);
 
 	for (bindex = bstart; bindex <= bend; bindex++) {
 		hidden_inode = itohi_index(inode, bindex);
@@ -900,9 +864,9 @@ static int unionfs_permission(struct inode *inode, int mask,
 		 * we dont have to check for files, if we are checking for
 		 * directories.
 		 */
-		if (!S_ISDIR(hidden_inode->i_mode) && (!is_file))
+		if (!is_file && !S_ISDIR(hidden_inode->i_mode))
 			continue;
-		/* We use our own special version of permission, such that 
+		/* We use our own special version of permission, such that
 		 * only the first branch returns -EROFS. */
 		err = inode_permission(hidden_inode, mask, nd, bindex);
 		/* The permissions are an intersection of the overall directory
@@ -910,12 +874,11 @@ static int unionfs_permission(struct inode *inode, int mask,
 		if (err)
 			goto out;
 		/* only the leftmost file matters. */
-		if (is_file) {
-			if (mask & MAY_WRITE) {
+		if (is_file || write_mask) {
+			if (is_file && write_mask) {
 				err = get_write_access(hidden_inode);
-				if (err)
-					goto out;
-				put_write_access(hidden_inode);
+				if (!err)
+					put_write_access(hidden_inode);
 			}
 			break;
 		}
@@ -978,15 +941,30 @@ static int unionfs_setattr(struct dentry *dentry, struct iattr *ia)
 			goto out;
 		break;
 	}
+#ifdef UNIONFS_MMAP
+	/*
+	 * SP: notify_change will change the lower file's size,
+	 * but we need to truncate the page tables, so need to call
+	 * vmtruncate()
+	 */
 
+	if (ia->ia_valid & ATTR_SIZE) {
+		if (ia->ia_size != i_size_read(inode)) {
+			err = vmtruncate(inode, ia->ia_size);
+			if (err) {
+				printk("unionfs_setattr: vmtruncate failed\n");
+			}
+		}
+	}
+#endif
 	/* get the size from the first hidden inode */
 	hidden_inode = itohi(dentry->d_inode);
-	fist_checkinode(inode, "unionfs_setattr");
+	checkinode(inode, "unionfs_setattr");
 	fist_copy_attr_all(inode, hidden_inode);
 
       out:
 	unlock_dentry(dentry);
-	fist_checkinode(inode, "post unionfs_setattr");
+	checkinode(inode, "post unionfs_setattr");
 	print_exit_status(err);
 	return err;
 }
