@@ -102,10 +102,9 @@ then
 	touch ${MOLROOT}/MOL/MOLFLAG && chkok || chkfail
 	echo "MOLVERSION=${MOLVERSION}"         >> ${MOLROOT}/MOL/MOLFLAG && chkok || chkfail
 	echo "KERNEL_RELEASE=${KERNEL_RELEASE}" >> ${MOLROOT}/MOL/MOLFLAG && chkok || chkfail
-	mkdir -p ${MOLROOT}/etc/mol  && chkok || chkfail # 0.6.0
-	echo "Image" > ${MOLROOT}/etc/mol/MOLTYPE && chkok || chkfail # 0.6.0
+	echo "MOL_TYPE=\"Image\""               >> ${MOLROOT}/MOL/MOLFLAG && chkok || chkfail # 0.6.1
 	mv -f ${BUILD_ROOT}/MOL ${MOLROOT}/MOL/MOL && chkok || chkfail
-	echo "Image" > ${MOLROOT}/.BASETYPE # 0.6.0
+	#echo "Image" > ${MOLROOT}/.BASETYPE # 0.6.0
 
 ## Raw system ##
 else
