@@ -1,8 +1,8 @@
 /*
- * Copyright (c) 2003-2005 Erez Zadok
  * Copyright (c) 2003-2006 Erez Zadok
  * Copyright (c) 2003-2006 Charles P. Wright
- * Copyright (c) 2005-2006 Josef Sipek
+ * Copyright (c) 2005-2006 Josef 'Jeff' Sipek
+ * Copyright (c) 2005-2006 Junjiro Okajima
  * Copyright (c) 2005      Arun M. Krishnakumar
  * Copyright (c) 2005-2006 David P. Quigley
  * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
@@ -17,7 +17,7 @@
  * This Copyright notice must be kept intact and distributed with all sources.
  */
 /*
- *  $Id: super.c,v 1.89 2006/02/20 22:23:01 dquigley Exp $
+ *  $Id: super.c,v 1.99 2006/08/25 18:53:03 jsipek Exp $
  */
 
 #include "unionfs.h"
@@ -28,7 +28,12 @@ static kmem_cache_t *unionfs_inode_cachep;
 
 static void unionfs_read_inode(struct inode *inode)
 {
+#ifdef UNIONFS_MMAP
+	/* SP: use real address operations */
+	extern struct address_space_operations unionfs_aops;
+#else
 	static struct address_space_operations unionfs_empty_aops;
+#endif
 	int size;
 
 	print_entry_location();
@@ -60,10 +65,14 @@ static void unionfs_read_inode(struct inode *inode)
 	inode->i_version++;
 	inode->i_op = &unionfs_main_iops;
 	inode->i_fop = &unionfs_main_fops;
+#ifdef UNIONFS_MMAP
+	inode->i_mapping->a_ops = &unionfs_aops;
+#else
 	/* I don't think ->a_ops is ever allowed to be NULL */
 	inode->i_mapping->a_ops = &unionfs_empty_aops;
-	fist_dprint(7, "setting inode 0x%p a_ops to empty (0x%p)\n",
-		    inode, inode->i_mapping->a_ops);
+	dprint(PRINT_DEBUG, "setting inode 0x%p a_ops to empty (0x%p)\n",
+	       inode, inode->i_mapping->a_ops);
+#endif
 
 	print_exit_location();
 }
@@ -71,19 +80,24 @@ static void unionfs_read_inode(struct inode *inode)
 static void unionfs_put_inode(struct inode *inode)
 {
 	print_entry_location();
-	fist_dprint(8, "%s i_count = %d, i_nlink = %d\n", __FUNCTION__,
-		    atomic_read(&inode->i_count), inode->i_nlink);
+
+	dprint(PRINT_DEBUG, "%s i_count = %d, i_nlink = %d\n", __FUNCTION__,
+	       atomic_read(&inode->i_count), inode->i_nlink);
+
 	/*
 	 * This is really funky stuff:
-	 * Basically, if i_count == 1, iput will then decrement it and this inode will be destroyed.
-	 * It is currently holding a reference to the hidden inode.
-	 * Therefore, it needs to release that reference by calling iput on the hidden inode.
-	 * iput() _will_ do it for us (by calling our clear_inode), but _only_ if i_nlink == 0.
-	 * The problem is, NFS keeps i_nlink == 1 for silly_rename'd files.
-	 * So we must for our i_nlink to 0 here to trick iput() into calling our clear_inode.
+	 * Basically, if i_count == 1, iput will then decrement it and this
+	 * inode will be destroyed.  It is currently holding a reference to the
+	 * hidden inode.  Therefore, it needs to release that reference by
+	 * calling iput on the hidden inode.  iput() _will_ do it for us (by
+	 * calling our clear_inode), but _only_ if i_nlink == 0.  The problem
+	 * is, NFS keeps i_nlink == 1 for silly_rename'd files.  So we must for
+	 * our i_nlink to 0 here to trick iput() into calling our clear_inode.
 	 */
+
 	if (atomic_read(&inode->i_count) == 1)
 		inode->i_nlink = 0;
+
 	print_exit_location();
 }
 
@@ -98,8 +112,16 @@ static void unionfs_delete_inode(struct inode *inode)
 {
 	print_entry_location();
 
-	fist_checkinode(inode, "unionfs_delete_inode IN");
+	checkinode(inode, "unionfs_delete_inode IN");
 	inode->i_size = 0;	/* every f/s seems to do that */
+
+#ifdef UNIONFS_MMAP
+	/* SP: if you try to clear_inode() when
+	 * inode->i_data.nrpages != 0, you'll hit a BUG
+	 * this is also what generic_delete_inode does */
+	if (inode->i_data.nrpages)
+		truncate_inode_pages(&inode->i_data, 0);
+#endif
 	clear_inode(inode);
 
 	print_exit_location();
@@ -114,8 +136,8 @@ static void unionfs_put_super(struct super_block *sb)
 	print_entry_location();
 
 	if ((spd = stopd(sb))) {
-		/* XXX: Free persistent inode stuff. */
 #ifdef UNIONFS_IMAP
+		/* XXX: Free persistent inode stuff. */
 		cleanup_imap_data(sb);
 #endif
 		bstart = sbstart(sb);
@@ -131,19 +153,22 @@ static void unionfs_put_super(struct super_block *sb)
 		KFREE(spd);
 		stopd_lhs(sb) = NULL;
 	}
-	fist_dprint(6, "unionfs: released super\n");
+	dprint(PRINT_DEBUG, "unionfs: released super\n");
 
 	print_exit_location();
 }
 
-static int unionfs_statfs(struct super_block *sb, struct kstatfs *buf)
+static int unionfs_statfs(struct dentry *dentry, struct kstatfs *buf)
 {
 	int err = 0;
-	struct super_block *hidden_sb;
+	struct super_block *sb, *hidden_sb;
 	struct kstatfs rsb;
 	int bindex, bindex1, bstart, bend;
 
 	print_entry_location();
+
+	sb = dentry->d_sb;
+
 	memset(buf, 0, sizeof(struct kstatfs));
 	buf->f_type = UNIONFS_SUPER_MAGIC;
 
@@ -168,11 +193,11 @@ static int unionfs_statfs(struct super_block *sb, struct kstatfs *buf)
 			continue;
 		}
 
-		err = vfs_statfs(hidden_sb, &rsb);
-		fist_dprint(8,
-			    "adding values for bindex:%d bsize:%d blocks:%d bfree:%d bavail:%d\n",
-			    bindex, (int)rsb.f_bsize, (int)rsb.f_blocks,
-			    (int)rsb.f_bfree, (int)rsb.f_bavail);
+		err = vfs_statfs(hidden_sb->s_root, &rsb);
+		dprint(PRINT_DEBUG,
+		       "adding values for bindex:%d bsize:%d blocks:%d bfree:%d bavail:%d\n",
+		       bindex, (int)rsb.f_bsize, (int)rsb.f_blocks,
+		       (int)rsb.f_bfree, (int)rsb.f_bavail);
 
 		if (!buf->f_frsize)
 			buf->f_frsize = rsb.f_frsize;
@@ -264,7 +289,7 @@ static void unionfs_clear_inode(struct inode *inode)
 
 	print_entry_location();
 
-	fist_checkinode(inode, "unionfs_clear_inode IN");
+	checkinode(inode, "unionfs_clear_inode IN");
 
 	list_for_each_safe(pos, n, &itopd(inode)->uii_readdircache) {
 		rdstate = list_entry(pos, struct unionfs_dir_state, uds_cache);
@@ -389,8 +414,6 @@ static int unionfs_write_inode(struct inode *inode, int sync)
 }
 
 /*
- * Called in do_umount() if the MNT_FORCE flag was used and this
- * function is defined.  See comment in linux/fs/super.c:do_umount().
  * Used only in nfs, to kill any pending RPC tasks, so that subsequent
  * code can actually succeed and won't leave tasks that need handling.
  *
@@ -398,21 +421,34 @@ static int unionfs_write_inode(struct inode *inode, int sync)
  * left in the kernel after a user level file server (such as amd)
  * dies.
  */
-static void unionfs_umount_begin(struct super_block *sb)
+static void unionfs_umount_begin(struct vfsmount *mnt, int flags)
 {
-	struct super_block *hidden_sb;
+	struct super_block *sb, *hidden_sb;
+	struct vfsmount *hidden_mnt;
 	int bindex, bstart, bend;
 
 	print_entry_location();
 
+	if (!(flags & MNT_FORCE))
+		/* we are not being MNT_FORCEd, therefore we should emulate old
+		 * behaviour
+		 */
+		goto out;
+
+	sb = mnt->mnt_sb;
+
 	bstart = sbstart(sb);
 	bend = sbend(sb);
 	for (bindex = bstart; bindex <= bend; bindex++) {
-		hidden_sb = stohs_index(sb, bindex);
-		if (hidden_sb && hidden_sb->s_op &&
+		hidden_mnt = stohiddenmnt_index(sb, bindex);
+		hidden_sb  = stohs_index(sb, bindex);
+
+		if (hidden_mnt && hidden_sb && hidden_sb->s_op &&
 		    hidden_sb->s_op->umount_begin)
-			hidden_sb->s_op->umount_begin(hidden_sb);
+			hidden_sb->s_op->umount_begin(hidden_mnt, flags);
 	}
+
+out:
 	print_exit_location();
 }
 
@@ -451,7 +487,7 @@ static int unionfs_show_options(struct seq_file *m, struct vfsmount *mnt)
 		}
 	}
 
-	seq_printf(m, ",debug=%d", fist_get_debug_value());
+	seq_printf(m, ",debug=%u", get_debug_mask());
 
 #ifdef UNIONFS_DELETE_ALL
 	if (IS_SET(sb, DELETE_ALL))
@@ -459,11 +495,6 @@ static int unionfs_show_options(struct seq_file *m, struct vfsmount *mnt)
 	else
 #endif
 		seq_printf(m, ",delete=whiteout");
-	if (IS_SET(sb, COPYUP_CURRENT_USER))
-		seq_printf(m, ",copyup=currentuser");
-	else
-		seq_printf(m, ",copyup=preserve");
-
       out:
 	if (tmp)
 		free_page(tmp);
@@ -471,25 +502,301 @@ static int unionfs_show_options(struct seq_file *m, struct vfsmount *mnt)
 	return ret;
 }
 
-/**
- * The entry given by dentry here is always a directory.
- * We can't just do dentry->d_parent, because it may
- * not be there, since this dentry could have been
- * created by calling d_alloc_anon.
+#ifdef CONFIG_EXPORTFS
+/*
+ * export operations.
+ * unionfs cannot handle disconnected dentry, since it has no hidden dentries.
  */
-static struct dentry *unionfs_get_parent(struct dentry *dentry)
-{
+/* un-tested 64 bit environment (pointer and inode number) */
 
-	struct dentry *ret;
-	struct inode *childinode;
-	childinode = dentry->d_inode;
-	ret = d_find_alias(childinode);
-	return ret;
+#define is_anon(d) ((d)->d_flags & DCACHE_DISCONNECTED)
+extern struct export_operations export_op_default;
+
+static void prepend_path(char **path, const char *name, int len)
+{
+	*path -= len;
+	memcpy(*path, name, len);
+	(*path)--;
+	**path = '/';
+}
+
+struct filldir_arg {
+	int found, called;
+	char *path;
+	ino_t ino, parent_ino;
+};
+
+static int filldir(void *arg, const char *name, int len, loff_t pos, ino_t ino,
+		   unsigned int d_type)
+{
+	struct filldir_arg *a = arg;
+
+	a->called++;
+	if (len == 2 && !strncmp(name, "..", 2)) {
+		a->parent_ino = ino;
+		a->found++;
+	} else if (ino == a->ino) {
+		if (len != 1 || *name != '.')
+			prepend_path(&a->path, name, len);
+		a->found++;
+	}
+	return (a->found == 2) ? 1 : 0;
+}
+
+static struct dentry *get_hidden_parent(struct super_block *hidden_sb,
+					ino_t hidden_parent_ino)
+{
+	__u32 fh[2];
+
+	if (hidden_sb->s_root->d_inode->i_ino == hidden_parent_ino)
+		return DGET(hidden_sb->s_root);
+
+	fh[0] = hidden_parent_ino;
+	fh[1] = 0;
+	return export_op_default.get_dentry(hidden_sb, fh);
+}
+
+static struct dentry *do_get_dentry(struct super_block *sb, ino_t ino,
+				    __u32 gen, struct dentry *hidden_root,
+				    ino_t hidden_ino, ino_t hidden_parent_ino)
+{
+	struct dentry *dentry, *hidden_parent, *parent;
+	char *path, *p;
+	struct filldir_arg arg = {
+		.ino = hidden_ino,
+		.parent_ino = hidden_parent_ino
+	};
+	int open_flags, err, bindex, bend, found;
+	struct file *hidden_file;
+	struct super_block *hidden_sb;
+
+	print_entry("hr%p, hi%lu, hpi%lu",
+		    hidden_root, hidden_ino, hidden_parent_ino);
+
+	dentry = ERR_PTR(-ENOMEM);
+	path = __getname();
+	if (!path)
+		goto out;
+	arg.path = path + PATH_MAX - 1;
+	*arg.path = 0;
+
+	open_flags = O_RDONLY | O_DIRECTORY /* | O_NOATIME */ ;
+	if (force_o_largefile())
+		open_flags |= O_LARGEFILE;
+
+	dentry = ERR_PTR(-ESTALE);
+	unionfs_read_lock(sb);
+	lock_dentry(sb->s_root);
+	bend = dbend(sb->s_root);
+	found = -1;
+	for (bindex = 0; found == -1 && bindex <= bend; bindex++)
+		if (hidden_root == dtohd_index(sb->s_root, bindex))
+			found = bindex;
+	unlock_dentry(sb->s_root);
+	if (found == -1)
+		goto out_unlock;
+
+	bindex = found;
+	hidden_sb = stohs_index(sb, bindex);
+	while (1) {
+		hidden_parent = get_hidden_parent(hidden_sb, hidden_parent_ino);
+		dentry = hidden_parent;
+		if (IS_ERR(hidden_parent))
+			goto out_unlock;
+
+		branchget(sb, bindex);
+		hidden_file = DENTRY_OPEN(DGET(hidden_parent), NULL,
+					  open_flags);
+		if (IS_ERR(hidden_file)) {
+			dentry = (void *)hidden_file;
+			DPUT(hidden_parent);
+			branchput(sb, bindex);
+			goto out_unlock;
+		}
+
+		arg.found = 0;
+		while (arg.found != 2) {
+			arg.called = 0;
+			err = vfs_readdir(hidden_file, filldir, &arg);
+			if (!arg.called || err < 0)
+				break;
+		}
+		fput(hidden_file);
+		branchput(sb, bindex);
+		if (arg.found != 2) {
+			dentry = ERR_PTR(-ESTALE);
+			DPUT(hidden_parent);
+			goto out_unlock;
+		}
+
+		DPUT(hidden_parent);
+		if (hidden_parent_ino == hidden_root->d_inode->i_ino)
+			break;
+		arg.ino = hidden_parent_ino;
+		hidden_parent_ino = arg.parent_ino;
+	}
+	BUG_ON(arg.path < path);
+
+	parent = DGET(sb->s_root);
+	p = strchr(++arg.path, '/');
+	while (p) {
+		mutex_lock(&parent->d_inode->i_mutex);
+		dentry = LOOKUP_ONE_LEN(arg.path, parent, p - arg.path);
+		mutex_unlock(&parent->d_inode->i_mutex);
+		DPUT(parent);
+		if (IS_ERR(dentry))
+			goto out_unlock;
+		if (!dentry->d_inode || !S_ISDIR(dentry->d_inode->i_mode)) {
+			DPUT(dentry);
+			dentry = ERR_PTR(-ESTALE);
+			goto out_unlock;
+		}
+		parent = dentry;
+		arg.path = p + 1;
+		p = strchr(arg.path, '/');
+	}
+	mutex_lock(&parent->d_inode->i_mutex);
+	dentry = LOOKUP_ONE_LEN(arg.path, parent, strlen(arg.path));
+	mutex_unlock(&parent->d_inode->i_mutex);
+	DPUT(parent);
+	if (!IS_ERR(dentry)
+	    && (!dentry->d_inode
+		|| dentry->d_inode->i_ino != ino
+		|| dentry->d_inode->i_generation != gen)) {
+		DPUT(dentry);
+		dentry = ERR_PTR(-ESTALE);
+	}
+
+      out_unlock:
+	unionfs_read_unlock(sb);
+	__putname(path);
+      out:
+	print_exit_pointer(dentry);
+	return dentry;
+}
+
+enum {
+	FhHead = 4, FhHRoot1 = FhHead, FhHRoot2,
+	FhHIno1, FhHIno2, FhHPIno1, FhHPIno2,
+	FhTail
+};
+
+static void do_decode(__u32 * fh, struct dentry **hidden_root,
+		      ino_t * hidden_ino, ino_t * hidden_parent_ino)
+{
+	unsigned long root;
+	
+	root = fh[FhHRoot2];
+	*hidden_ino = fh[FhHIno2];
+	*hidden_parent_ino = fh[FhHPIno2];
+#if BITS_PER_LONG == 64
+	root |= ((unsigned long)fh[FhHRoot1]) << 32;
+	*hidden_ino |= ((unsigned long) fh[FhHIno1]) << 32;
+	*hidden_parent_ino |= ((unsigned long) fh[FhHPIno1]) << 32;
+#elif BITS_PER_LONG == 32
+	/* ok */
+#else
+#error unknown size
+#endif
+
+	*hidden_root = (struct dentry*) root;
+}
+
+static int unionfs_encode_fh(struct dentry *dentry, __u32 * fh, int *max_len,
+			     int connectable)
+{
+	int type, len, bindex;
+	struct super_block *sb;
+	struct dentry *h_root;
+	ino_t h_ino, hp_ino;
+	static int warn;
+
+	print_entry("dentry %p", dentry);
+	BUG_ON(is_anon(dentry) || !dentry->d_inode
+	       || is_anon(dentry->d_parent));
+
+#ifdef UNIONFS_IMAP
+	if (!warn && stopd(dentry->d_sb)->usi_persistent)
+		warn++;
+#endif
+	if (!warn) {
+		printk(KERN_WARNING "Exporting Unionfs without imap"
+		       " option may stop your NFS server or client");
+		warn++;
+	}
+
+	sb = dentry->d_sb;
+	unionfs_read_lock(sb);
+	lock_dentry(dentry);
+
+	len = *max_len;
+	type = export_op_default.encode_fh(dentry, fh, max_len, connectable);
+	if (type == 255 || *max_len > FhHead || len < FhTail) {
+		type = 255;
+		goto out;
+	}
+
+	*max_len = FhTail;
+	bindex = dbstart(dentry);
+	lock_dentry(sb->s_root);
+	h_root = dtohd_index(sb->s_root, bindex);
+	unlock_dentry(sb->s_root);
+	h_ino = itohi_index(dentry->d_inode, bindex)->i_ino;
+	hp_ino = parent_ino(dtohd(dentry));
+	fh[FhHRoot2] = (unsigned long) h_root;
+	fh[FhHIno2] = h_ino;
+	fh[FhHPIno2] = hp_ino;
+#if BITS_PER_LONG == 64
+	fh[FhHRoot1] = ((unsigned long) h_root) >> 32;
+	fh[FhHIno1] = h_ino >> 32;
+	fh[FhHPIno1] = hp_ino >> 32;
+#endif
+
+      out:
+	unionfs_print(PRINT_MAIN_EXIT, "%d, fh{i%u, g%d, hr%x, hi%u, hpi%u}\n",
+		      type, fh[0], fh[1], fh[FhHRoot2], fh[FhHIno2],
+		      fh[FhHPIno2]);
+	unlock_dentry(dentry);
+	unionfs_read_unlock(sb);
+	return type;
+}
+
+static struct dentry *unionfs_decode_fh(struct super_block *sb, __u32 * fh,
+					int fh_len, int fh_type,
+					int (*acceptable) (void *context,
+							   struct dentry * de),
+					void *context)
+{
+	struct dentry *dentry, *hidden_root;
+	ino_t hidden_ino, hidden_parent_ino;
+
+	print_entry("%d, fh{i%u, g%d, hr%x, hi%u, hpi%u}",
+		    fh_type, fh[0], fh[1], fh[FhHRoot2], fh[FhHIno2],
+		    fh[FhHPIno2]);
+
+	dentry = export_op_default.get_dentry(sb, fh);
+	if (!dentry || IS_ERR(dentry) || (dentry->d_inode && !is_anon(dentry)))
+		return dentry;
+
+	d_drop(dentry);
+	DPUT(dentry);
+	do_decode(fh, &hidden_root, &hidden_ino, &hidden_parent_ino);
+	dentry = do_get_dentry(sb, fh[0], fh[1], hidden_root, hidden_ino,
+			       hidden_parent_ino);
+	if (!IS_ERR(dentry)) {
+		if (acceptable(context, dentry))
+			return dentry;	/* success */
+		DPUT(dentry);
+		dentry = NULL;
+	}
+	return dentry;
 }
 
 struct export_operations unionfs_export_ops = {
-	.get_parent = unionfs_get_parent
+	.decode_fh = unionfs_decode_fh,
+	.encode_fh = unionfs_encode_fh
 };
+#endif
 
 struct super_operations unionfs_sops = {
 	.read_inode = unionfs_read_inode,
@@ -504,9 +811,6 @@ struct super_operations unionfs_sops = {
 	.write_inode = unionfs_write_inode,
 	.alloc_inode = unionfs_alloc_inode,
 	.destroy_inode = unionfs_destroy_inode,
-#ifdef SPLIT_VIEW_CACHES
-	.select_super = unionfs_select_super,
-#endif
 };
 
 /*
