@@ -45,7 +45,7 @@
 #include <linux/poll.h>
 
 #ifndef UNIONFS_UNSUPPORTED
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
+#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)) || (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15))
 #warning You are compiling Unionfs on an unsupported kernel version.
 #warning To compile Unionfs, you will need to define UNIONFS_UNSUPPORTED.
 #warning Try adding: EXTRACFLAGS=-DUNIONFS_UNSUPPORTED to fistdev.mk
@@ -388,8 +388,6 @@ extern int unionfs_partial_lookup(struct dentry *dentry);
 /* Pass an unionfs dentry and an index and it will try to create a whiteout in branch 'index'.
    On error, it will proceed to a branch to the left */
 extern int create_whiteout(struct dentry *dentry, int start);
-extern int create_whiteout_parent(struct dentry *parent_dentry,
-				  const char *filename, int start);
 /* copies a file from dbstart to newbindex branch */
 extern int copyup_file(struct inode *dir, struct file *file, int bstart,
 		       int newbindex, loff_t size);
@@ -485,10 +483,6 @@ extern ssize_t unionfs_listxattr(struct dentry *dentry, char *list,
 int unionfs_setxattr(struct dentry *dentry, const char *name, const void *value,
 		     size_t size, int flags);
 
-#define copy_inode_size(dst, src) \
-    dst->i_size = src->i_size; \
-    dst->i_blocks = src->i_blocks;
-
 /* The root directory is unhashed, but isn't deleted. */
 static inline int d_deleted(struct dentry *d)
 {
@@ -546,13 +540,14 @@ static inline void fist_copy_attr_timesizes(struct inode *dest,
 	dest->i_atime = src->i_atime;
 	dest->i_mtime = src->i_mtime;
 	dest->i_ctime = src->i_ctime;
-	copy_inode_size(dest, src);
+	dest->i_size = src->i_size;
+	dest->i_blocks = src->i_blocks;
 }
 static inline void fist_copy_attr_all(struct inode *dest,
 				      const struct inode *src)
 {
-
 	print_entry_location();
+
 	dest->i_mode = src->i_mode;
 	/* we do not need to copy if the file is a deleted file */
 	if (dest->i_nlink > 0)
@@ -565,10 +560,8 @@ static inline void fist_copy_attr_all(struct inode *dest,
 	dest->i_ctime = src->i_ctime;
 	dest->i_blksize = src->i_blksize;
 	dest->i_blkbits = src->i_blkbits;
-	copy_inode_size(dest, src);
-
-	//DQ: This was a change I noticed in the templates. In 2.6 they removedi_attr_flags.
-	//Which makes me think they rolled it into flags.
+	dest->i_size = src->i_size;
+	dest->i_blocks = src->i_blocks;
 	dest->i_flags = src->i_flags;
 
 	print_exit_location();
